@@ -32,9 +32,15 @@ typedef enum {
 	SOUP_MESSAGE_IO_STATE_CHUNK,
 	SOUP_MESSAGE_IO_STATE_CHUNK_END,
 	SOUP_MESSAGE_IO_STATE_TRAILERS,
+	SOUP_MESSAGE_IO_STATE_FINISHING,
 	SOUP_MESSAGE_IO_STATE_DONE
 } SoupMessageIOState;
 
+#define SOUP_MESSAGE_IO_STATE_ACTIVE(state) \
+	(state != SOUP_MESSAGE_IO_STATE_NOT_STARTED && \
+	 state != SOUP_MESSAGE_IO_STATE_BLOCKING && \
+	 state != SOUP_MESSAGE_IO_STATE_DONE)
+
 typedef struct {
 	SoupSocket           *sock;
 	SoupConnection       *conn;
@@ -66,8 +72,8 @@ typedef struct {
  */
 #define dummy_to_make_emacs_happy {
 #define SOUP_MESSAGE_IO_PREPARE_FOR_CALLBACK { gboolean cancelled; g_object_ref (msg);
-#define SOUP_MESSAGE_IO_RETURN_IF_CANCELLED_OR_PAUSED cancelled = (priv->io_data != io); g_object_unref (msg); if (cancelled || !io->read_tag || !io->write_tag) return; }
-#define SOUP_MESSAGE_IO_RETURN_VAL_IF_CANCELLED_OR_PAUSED(val) cancelled = (priv->io_data != io); g_object_unref (msg); if (cancelled || !io->read_tag || !io->write_tag) return val; }
+#define SOUP_MESSAGE_IO_RETURN_IF_CANCELLED_OR_PAUSED cancelled = (priv->io_data != io); g_object_unref (msg); if (cancelled || (!io->read_tag && !io->write_tag)) return; }
+#define SOUP_MESSAGE_IO_RETURN_VAL_IF_CANCELLED_OR_PAUSED(val) cancelled = (priv->io_data != io); g_object_unref (msg); if (cancelled || (!io->read_tag && !io->write_tag)) return val; }
 
 #define RESPONSE_BLOCK_SIZE 8192
 
@@ -75,13 +81,15 @@ static void
 io_cleanup (SoupMessage *msg)
 {
 	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
-	SoupMessageIOData *io = priv->io_data;
-
-	if (!io)
-		return;
+	SoupMessageIOData *io;
 
 	soup_message_io_stop (msg);
 
+	io = priv->io_data;
+	if (!io)
+		return;
+	priv->io_data = NULL;
+
 	if (io->sock)
 		g_object_unref (io->sock);
 	if (io->conn)
@@ -94,7 +102,6 @@ io_cleanup (SoupMessage *msg)
 	g_string_free (io->write_buf, TRUE);
 
 	g_free (io);
-	priv->io_data = NULL;
 }
 
 /**
@@ -126,10 +133,14 @@ soup_message_io_stop (SoupMessage *msg)
 		io->err_tag = 0;
 	}
 
-	if (io->read_state != SOUP_MESSAGE_IO_STATE_DONE)
+	if (io->read_state < SOUP_MESSAGE_IO_STATE_FINISHING)
 		soup_socket_disconnect (io->sock);
-	else if (io->conn)
-		soup_connection_release (io->conn);
+	else if (io->conn) {
+		SoupConnection *conn = io->conn;
+		io->conn = NULL;
+		soup_connection_release (conn);
+		g_object_unref (conn);
+	}
 }
 
 #define SOUP_MESSAGE_IO_EOL            "\r\n"
@@ -149,6 +160,7 @@ soup_message_io_finished (SoupMessage *msg)
 	g_object_unref (msg);
 }
 
+static void io_read (SoupSocket *sock, SoupMessage *msg);
 
 static void
 io_error (SoupSocket *sock, SoupMessage *msg)
@@ -159,8 +171,8 @@ io_error (SoupSocket *sock, SoupMessage *msg)
 	/* Closing the connection to signify EOF is sometimes ok */
 	if (io->read_state == SOUP_MESSAGE_IO_STATE_BODY &&
 	    io->read_encoding == SOUP_TRANSFER_UNKNOWN) {
-		io->read_state = SOUP_MESSAGE_IO_STATE_DONE;
-		soup_message_io_finished (msg);
+		io->read_state = SOUP_MESSAGE_IO_STATE_FINISHING;
+		io_read (sock, msg);
 		return;
 	}
 
@@ -179,13 +191,27 @@ io_error (SoupSocket *sock, SoupMessage *msg)
 	soup_message_io_finished (msg);
 }
 
+/* Reads data from io->sock into io->read_meta_buf up until @boundary.
+ * (This function is used to read metadata, and read_body_chunk() is
+ * used to read the message body contents.)
+ *
+ * read_metadata, read_body_chunk, and write_data all use the same
+ * convention for return values: if they return %TRUE, it means
+ * they've completely finished the requested read/write, and the
+ * caller should move on to the next step. If they return %FALSE, it
+ * means that either (a) the socket returned SOUP_SOCKET_WOULD_BLOCK,
+ * so the caller should give up for now and wait for the socket to
+ * emit a signal, or (b) the socket returned an error, and io_error()
+ * was called to process it and cancel the I/O. So either way, if the
+ * function returns %FALSE, the caller should return immediately.
+ */
 static gboolean
 read_metadata (SoupMessage *msg, const char *boundary)
 {
 	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
 	SoupMessageIOData *io = priv->io_data;
 	SoupSocketIOStatus status;
-	char read_buf[RESPONSE_BLOCK_SIZE];
+	guchar read_buf[RESPONSE_BLOCK_SIZE];
 	guint boundary_len = strlen (boundary);
 	gsize nread;
 	gboolean done;
@@ -213,13 +239,23 @@ read_metadata (SoupMessage *msg, const char *boundary)
 	return TRUE;
 }
 
+/* Reads as much message body data as is available on io->sock (but no
+ * further than the end of the current message body or chunk). On a
+ * successful read, emits "got_chunk" (possibly multiple times), and
+ * if io->read_buf is non-%NULL (meaning that the message doesn't have
+ * %SOUP_MESSAGE_OVERWRITE_CHUNKS set), the data will be appended to
+ * it.
+ *
+ * See the note at read_metadata() for an explanation of the return
+ * value.
+ */
 static gboolean
 read_body_chunk (SoupMessage *msg)
 {
 	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
 	SoupMessageIOData *io = priv->io_data;
 	SoupSocketIOStatus status;
-	char read_buf[RESPONSE_BLOCK_SIZE];
+	guchar read_buf[RESPONSE_BLOCK_SIZE];
 	guint len = sizeof (read_buf);
 	gboolean read_to_eof = (io->read_encoding == SOUP_TRANSFER_UNKNOWN);
 	gsize nread;
@@ -236,7 +272,7 @@ read_body_chunk (SoupMessage *msg)
 				break;
 
 			io->read_body->owner  = SOUP_BUFFER_STATIC;
-			io->read_body->body   = read_buf;
+			io->read_body->body   = (char *)read_buf;
 			io->read_body->length = nread;
 
 			SOUP_MESSAGE_IO_PREPARE_FOR_CALLBACK;
@@ -267,6 +303,9 @@ read_body_chunk (SoupMessage *msg)
 	return TRUE;
 }
 
+/* Attempts to write @len bytes from @data. See the note at
+ * read_metadata() for an explanation of the return value.
+ */
 static gboolean
 write_data (SoupMessage *msg, const char *data, guint len)
 {
@@ -308,8 +347,6 @@ io_body_state (SoupTransferEncoding encoding)
 		return SOUP_MESSAGE_IO_STATE_BODY;
 }
 
-static void io_read (SoupSocket *sock, SoupMessage *msg);
-
 /*
  * There are two request/response formats: the basic request/response,
  * possibly with one or more unsolicited informational responses (such
@@ -369,6 +406,7 @@ io_write (SoupSocket *sock, SoupMessage *msg)
 				/* Stop and wait for the body now */
 				io->write_state =
 					SOUP_MESSAGE_IO_STATE_BLOCKING;
+				io->read_state = io_body_state (io->read_encoding);
 			} else {
 				/* We just wrote a 1xx response
 				 * header, so stay in STATE_HEADERS.
@@ -412,7 +450,7 @@ io_write (SoupSocket *sock, SoupMessage *msg)
 				 io->write_body->length))
 			return;
 
-		io->write_state = SOUP_MESSAGE_IO_STATE_DONE;
+		io->write_state = SOUP_MESSAGE_IO_STATE_FINISHING;
 
 		SOUP_MESSAGE_IO_PREPARE_FOR_CALLBACK;
 		soup_message_wrote_body (msg);
@@ -485,7 +523,7 @@ io_write (SoupSocket *sock, SoupMessage *msg)
 				 SOUP_MESSAGE_IO_EOL_LEN))
 			return;
 
-		io->write_state = SOUP_MESSAGE_IO_STATE_DONE;
+		io->write_state = SOUP_MESSAGE_IO_STATE_FINISHING;
 
 		SOUP_MESSAGE_IO_PREPARE_FOR_CALLBACK;
 		soup_message_wrote_body (msg);
@@ -493,7 +531,13 @@ io_write (SoupSocket *sock, SoupMessage *msg)
 		/* fall through */
 
 
-	case SOUP_MESSAGE_IO_STATE_DONE:
+	case SOUP_MESSAGE_IO_STATE_FINISHING:
+		if (io->write_tag) {
+			g_signal_handler_disconnect (io->sock, io->write_tag);
+			io->write_tag = 0;
+		}
+		io->write_state = SOUP_MESSAGE_IO_STATE_DONE;
+
 		if (io->mode == SOUP_MESSAGE_IO_CLIENT) {
 			io->read_state = SOUP_MESSAGE_IO_STATE_HEADERS;
 			io_read (sock, msg);
@@ -502,6 +546,7 @@ io_write (SoupSocket *sock, SoupMessage *msg)
 		return;
 
 
+	case SOUP_MESSAGE_IO_STATE_DONE:
 	default:
 		g_return_if_reached ();
 	}
@@ -528,7 +573,7 @@ io_read (SoupSocket *sock, SoupMessage *msg)
 
 		io->read_meta_buf->len -= SOUP_MESSAGE_IO_EOL_LEN;
 		io->read_meta_buf->data[io->read_meta_buf->len] = '\0';
-		status = io->parse_headers_cb (msg, io->read_meta_buf->data,
+		status = io->parse_headers_cb (msg, (char *)io->read_meta_buf->data,
 					       io->read_meta_buf->len,
 					       &io->read_encoding,
 					       &io->read_length,
@@ -546,7 +591,7 @@ io_read (SoupSocket *sock, SoupMessage *msg)
 			soup_message_set_status (msg, status);
 			soup_message_add_header (msg->request_headers,
 						 "Connection", "close");
-			io->read_state = SOUP_MESSAGE_IO_STATE_DONE;
+			io->read_state = SOUP_MESSAGE_IO_STATE_FINISHING;
 			break;
 		}
 
@@ -566,13 +611,16 @@ io_read (SoupSocket *sock, SoupMessage *msg)
 		} else if (io->mode == SOUP_MESSAGE_IO_SERVER &&
 			   (priv->msg_flags & SOUP_MESSAGE_EXPECT_CONTINUE)) {
 			/* The client requested a Continue response. */
+			soup_message_set_status (msg, SOUP_STATUS_CONTINUE);
+			
 			io->write_state = SOUP_MESSAGE_IO_STATE_HEADERS;
 			io->read_state = SOUP_MESSAGE_IO_STATE_BLOCKING;
 		} else
 			io->read_state = io_body_state (io->read_encoding);
 
 		SOUP_MESSAGE_IO_PREPARE_FOR_CALLBACK;
-		if (SOUP_STATUS_IS_INFORMATIONAL (msg->status_code)) {
+		if (SOUP_STATUS_IS_INFORMATIONAL (msg->status_code) &&
+		    !(priv->msg_flags & SOUP_MESSAGE_EXPECT_CONTINUE)) {
 			soup_message_got_informational (msg);
 			soup_message_cleanup_response (msg);
 		} else
@@ -595,14 +643,14 @@ io_read (SoupSocket *sock, SoupMessage *msg)
 	got_body:
 		if (io->read_buf) {
 			io->read_body->owner = SOUP_BUFFER_SYSTEM_OWNED;
-			io->read_body->body = io->read_buf->data;
+			io->read_body->body = (char *)io->read_buf->data;
 			io->read_body->length = io->read_buf->len;
 
 			g_byte_array_free (io->read_buf, FALSE);
 			io->read_buf = NULL;
 		}
 
-		io->read_state = SOUP_MESSAGE_IO_STATE_DONE;
+		io->read_state = SOUP_MESSAGE_IO_STATE_FINISHING;
 
 		SOUP_MESSAGE_IO_PREPARE_FOR_CALLBACK;
 		soup_message_got_body (msg);
@@ -614,7 +662,7 @@ io_read (SoupSocket *sock, SoupMessage *msg)
 		if (!read_metadata (msg, SOUP_MESSAGE_IO_EOL))
 			return;
 
-		io->read_length = strtoul (io->read_meta_buf->data, NULL, 16);
+		io->read_length = strtoul ((char *)io->read_meta_buf->data, NULL, 16);
 		g_byte_array_set_size (io->read_meta_buf, 0);
 
 		if (io->read_length > 0)
@@ -653,7 +701,13 @@ io_read (SoupSocket *sock, SoupMessage *msg)
 		break;
 
 
-	case SOUP_MESSAGE_IO_STATE_DONE:
+	case SOUP_MESSAGE_IO_STATE_FINISHING:
+		if (io->read_tag) {
+			g_signal_handler_disconnect (io->sock, io->read_tag);
+			io->read_tag = 0;
+		}
+		io->read_state = SOUP_MESSAGE_IO_STATE_DONE;
+
 		if (io->mode == SOUP_MESSAGE_IO_SERVER) {
 			io->write_state = SOUP_MESSAGE_IO_STATE_HEADERS;
 			io_write (sock, msg);
@@ -662,6 +716,7 @@ io_read (SoupSocket *sock, SoupMessage *msg)
 		return;
 
 
+	case SOUP_MESSAGE_IO_STATE_DONE:
 	default:
 		g_return_if_reached ();
 	}
@@ -815,14 +870,34 @@ soup_message_io_unpause (SoupMessage *msg)
 	if (io->write_tag || io->read_tag)
 		return;
 
-	io->write_tag = g_signal_connect (io->sock, "writable",
-					  G_CALLBACK (io_write), msg);
-	io->read_tag = g_signal_connect (io->sock, "readable",
-					 G_CALLBACK (io_read), msg);
+	if (io->write_state != SOUP_MESSAGE_IO_STATE_DONE) {
+		io->write_tag = g_signal_connect (io->sock, "writable",
+						  G_CALLBACK (io_write), msg);
+	}
 
-	if (io->write_state != SOUP_MESSAGE_IO_STATE_NOT_STARTED &&
-	    io->write_state != SOUP_MESSAGE_IO_STATE_BLOCKING)
+	if (io->read_state != SOUP_MESSAGE_IO_STATE_DONE) {
+		io->read_tag = g_signal_connect (io->sock, "readable",
+						 G_CALLBACK (io_read), msg);
+	}
+
+	if (SOUP_MESSAGE_IO_STATE_ACTIVE (io->write_state))
 		io_write (io->sock, msg);
-	else
+	else if (SOUP_MESSAGE_IO_STATE_ACTIVE (io->read_state))
 		io_read (io->sock, msg);
 }
+
+/**
+ * soup_message_io_in_progress:
+ * @msg: a #SoupMessage
+ *
+ * Tests whether or not I/O is currently in progress on @msg.
+ *
+ * Return value: whether or not I/O is currently in progress.
+ **/
+gboolean
+soup_message_io_in_progress (SoupMessage *msg)
+{
+	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
+
+	return priv->io_data != NULL;
+}
